from ._GraphObj import GraphObj
from .graph_style.EdgeStyle import EdgeStyle
from .Node import Node


class Edge(GraphObj):
	def __init__(
			self, graph, start, end, id=None, value=None, label=None, style=None,
			**kwargs
	):
		"""
		:param Graph graph: a graph
		:param Node start: start node
		:param Node end: end node
		:param str or int id: an extra id (for when there are multiple edges between the same start and end nodes)
		:param value: some value the edge keeps
		:param str label: a label to show
		:param EdgeStyle or NoneType style: the style of the edge
		"""
		if style is None:
			style = EdgeStyle()
		elif isinstance(style, dict):
			style = EdgeStyle(**style)

		super().__init__(
			graph=graph, id=None, value=value, label=label, style=style,
			**kwargs
		)
		self._start = start
		self._end = end
		self.raw_id = id
		start.append_outward_edge(edge=self)
		end.append_inward_edge(edge=self)

	@property
	def style(self):
		"""
		:rtype: EdgeStyle
		"""
		return self._style

	@style.setter
	def style(self, style):
		"""
		:type style: EdgeStyle or dict
		"""
		if isinstance(style, dict):
			style = EdgeStyle(**style)
		self._style = style
		self._style_is_native = style is not None

	def is_similar_to(self, other):
		"""
		:type other: Edge
		:rtype: bool
		"""
		if not super().is_similar_to(other=other):
			return False
		return self._start.id == other._start.id and self._end.id == other._end.id and self.raw_id == other.raw_id

	def __getstate__(self):
		state = super().__getstate__()
		state.update({
			'start': None,
			'end': None,
			'id': self.raw_id
		})
		return state

	def __setstate__(self, state):
		super().__setstate__(state=state)
		self._start = state['start']
		self._end = state['end']
		self.raw_id = state['id']

	@property
	def id(self):
		if self._start is None:
			raise ValueError('This Edge does not have a start!')
		elif self._end is None:
			raise ValueError('This Edge does not have an end!')

		return self.start.id, self.end.id, self.raw_id

	def __str__(self):
		return f'Edge:{self.start}-{self.end}'

	def __repr__(self):
		return str(self)

	def get_graphviz_style_str(self):
		style = self.style
		if style is None:
			return ''
		else:
			return style.get_graphviz_str()

	def get_graphviz_label_str(self):
		if self.label_or_value is None:
			return ''
		else:
			return f'label="{self.label_or_value}"'

	def get_graphviz_str(self):
		without_style = f'"{self.start.name}" -> "{self.end.name}"'
		style = self.get_graphviz_style_str()
		label = self.get_graphviz_label_str()
		if style == '' and label == '':
			return without_style
		elif style == '':
			return f'{without_style} [{label}]'
		elif label == '':
			return f'{without_style} [{style}]'
		else:
			return f'{without_style} [{label} {style}]'

	@property
	def start(self):
		"""
		:rtype: Node
		"""
		if self._start is None:
			raise ValueError('Edge does not have a start!')
		return self._start

	@property
	def end(self):
		"""
		:rtype: Node
		"""
		if self._end is None:
			raise ValueError('Edge does not have an end!')
		return self._end

	def remove_self(self):
		if self._start is None or self._end is None:
			raise ValueError('Either start or end Node is missing!')
		self.start.remove_outward_edge(edge_id=self.id)
		self.end.remove_inward_edge(edge_id=self.id)
		self._start = None
		self._end = None

	def is_in_loop(self):
		return self.start.is_in_loop_with(self.end)

	'''
	@property
	def graph_style(self):
		return self.graph.edge_style
	'''



from .NodeStyle import NodeStyle
from .EdgeStyle import EdgeStyle, DEFAULT_FONT


def inherit_node_style_from_parent(node):
	"""
	:type node: abstract.Node.Node
	:rtype: NodeStyle
	"""
	scheme = node.graph.colour_scheme
	style = None
	colour = None

	if node.has_parents():
		best_candidates = []
		for parent in node.parents:
			if not parent.is_in_loop() or not parent.is_style_callable():
				style = parent.compiled_style.copy()
				best_candidates.append(style)

		if len(best_candidates) > 0:
			best_candidates.sort(key=lambda x: x.colour.usage_amount)
			style = best_candidates[0]
			colour = style.colour

	if style is None or colour is None:
		colour = scheme.least_used_colour
		style = NodeStyle(colour=colour)

	return style


def inherit_node_style_from_child(node):
	"""
	:type node: abstract.Node.Node
	:rtype: NodeStyle
	"""
	scheme = node.graph.colour_scheme
	style = None
	colour = None
	if node.has_children():
		best_candidates = []
		for child in node.children:
			if not child.is_in_loop() or not child.is_style_callable():
				style = child.compiled_style.copy()
				best_candidates.append(style)
		if len(best_candidates) > 0:
			best_candidates.sort(key=lambda x: x.colour.usage_amount)
			style = best_candidates[0]
			colour = style.colour

	if style is None or colour is None:
		colour = scheme.least_used_colour
		style = NodeStyle(colour=colour)

	return style


def inherit_edge_style_from_start(edge, font=DEFAULT_FONT):
	"""
	:type edge: abstract.Edge.Edge
	:rtype: EdgeStyle
	"""
	colour = edge.start.compiled_style.border_colour
	return EdgeStyle(colour=colour, font=font)


def inherit_edge_style_from_end(edge, font=DEFAULT_FONT):
	"""
	:type edge: abstract.Edge.Edge
	:rtype: EdgeStyle
	"""
	colour = edge.end.compiled_style.border_colour
	return EdgeStyle(colour=colour, font=font)








from .EdgeStyle import EdgeStyle
from .NodeStyle import NodeStyle
from .GraphStyle import GraphStyle
from .StyleMaster import StyleMaster, NodeStylist, EdgeStylist
from .RootStylist import RootStylist
from .BranchStylist import BranchStylist
from .RootAndBranchStylist import RootAndBranchStylist
from .ColourInverter import ColourInverter, NodeColourInverter, EdgeColourInverter


from .. import Graph
from copy import deepcopy
from .NodeStyle import NodeStyle


class SimpleNodeStylist:
	def __init__(self, style, condition=None):
		"""
		:type style: callable or NodeStyle
		:type condition: callable or NoneType
		"""
		self._condition = condition
		self._style = style

		if condition is None:
			def condition_function(node):
				return True
		elif not callable(condition):
			raise TypeError(f'condition should either be None or a function, but it is a {type(condition)}')
		else:
			condition_function = condition

		self._condition_function = condition_function

		if callable(style):
			def paint_node(node):
				style(node)

		else:
			def paint_node(node):
				node.style = style

		self._paint_node = paint_node

	def paint(self, graph):
		"""
		:type graph: Graph
		:rtype Graph
		"""
		for node in graph.nodes:
			if self._condition_function(node):
				self._paint_node(node=node)

		return graph

	def copy(self):
		"""
		:rtype: SimpleNodeStylist
		"""
		return self.__class__(style=deepcopy(self._style), condition=deepcopy(self._condition))



from .StyleMaster import StyleMaster
from .RootStylist import RootStylist
from .BranchStylist import BranchStylist
from .LoopStylist import LoopStylist
from .NodeStyle import NodeStyle


class RootAndBranchStylist(StyleMaster):
	def __init__(
			self, colour_scheme=None, node_style=None, font=None, node_shape=None,
			node_shape_style=None, edge_style=None, loop_node_style=None, loop_edge_style=None
	):
		self._colour_scheme = colour_scheme
		self._font = font
		self._node_shape = node_shape
		self._node_shape_style = node_shape_style
		self._node_style = node_style or NodeStyle(
			font=self._font, shape=self._node_shape, shape_style=self._node_shape_style
		)

		self._edge_style = edge_style

		super().__init__(
			stylists=[
				RootStylist(colour_scheme=self._colour_scheme, node_style=self._node_style),
				BranchStylist(edge_style=edge_style),
				LoopStylist(node_style=loop_node_style, edge_style=loop_edge_style)
			]
		)

	def copy(self):
		return self.__class__(
			colour_scheme=self._colour_scheme,
			node_style=self._node_style,
			font=self._font,
			node_shape=self._node_shape,
			node_shape_style=self._node_shape_style,
			edge_style=self._edge_style
		)



from colouration import Colour


STYLE_RENAME = {
	'colour': 'color',
	'fill_colour': 'fillcolor',
	'text_colour': 'fontcolor',
	'text_size': 'fontsize',
	'font': 'fontname',
	# 'label_style': 'lblstyle',
	'arrow_size': 'arrowsize'

}


class GraphObjStyle:
	def __init__(self, name=None, opacity=None, **kwargs):
		self._name = name,
		self._opacity = opacity
		self._dictionary = kwargs
		self.update_dictionary_keys()

	def update_dictionary_keys(self):
		for old_key, new_key in STYLE_RENAME.items():
			if old_key in self.dictionary:
				self._dictionary[new_key] = self._dictionary.pop(old_key)

	def __getstate__(self):
		return {
			'name': self._name,
			'dictionary': self._dictionary
		}

	def __setstate__(self, state):
		self._name = state['name']
		self._dictionary = state['dictionary']
		self.update_dictionary_keys()

	@property
	def name(self):
		"""
		:rtype: str or NoneType
		"""
		if self._name is not None:
			return str(self._name)
		else:
			return None

	def __eq__(self, other):
		"""
		:type other: GraphObjStyle
		:rtype: bool
		"""
		return self._dictionary == other._dictionary and self._name == other._name

	def __ne__(self, other):
		"""
		:type other: GraphObjStyle
		:rtype: bool
		"""
		return self._dictionary != other._dictionary and self._name == other._name

	@property
	def dictionary(self):
		return self._dictionary.copy()

	def copy(self):
		result = self.__class__(name=self._name)
		result._dictionary = self.dictionary.copy()
		return result

	def update(self, other, inplace=False):
		"""
		:param GraphObjStyle or dict other: another style or dictionary that would update this one
		:param bool inplace: if this object should be updated or a new one created
		:rtype: NoneType or GraphObjStyle
		"""
		if inplace:
			if isinstance(other, GraphObjStyle):
				self._dictionary.update(other.dictionary)
			elif isinstance(other, dict):
				self._dictionary.update(other)
		else:
			result = self.copy()
			result.update(other=other, inplace=True)
			return result

	def set(self, **kwargs):
		d = kwargs
		for old_key, new_key in STYLE_RENAME.items():
			if old_key in d:
				d[new_key] = d.pop(old_key)
		self.update(d, inplace=True)

	def get(self, item):
		if item in STYLE_RENAME:
			item = STYLE_RENAME[item]
		return self.dictionary[item]

	@property
	def colour(self):
		return self.get(item='colour')

	def get_graphviz_str(self):
		"""
		:return:
		"""
		key_values = []
		for key, value in self._dictionary.items():
			if isinstance(value, Colour):
				value.use()
				value = value.get_hexadecimal(opacity=self._opacity)

			if isinstance(key, str):
				key = key.strip('"')

			if isinstance(value, str):
				value = value.strip('"')

			if value is not None:
				key_values.append((key, value))

		return ' '.join([f'"{key}"="{value}"' for key, value in key_values])
